""" Constant related to Schemas """

# cookiecutter template parameter
SCHEMAS_REGISTRY = "AWS_Schema_registry"
SCHEMA_NAME = "AWS_Schema_name"
SCHEMA_ROOT = "AWS_Schema_root"
EVENT_BRIDGE_SOURCE = "AWS_Schema_source"
EVENT_BRIDGE_SOURCE_DETAIL_TYPE = "AWS_Schema_detail_type"

# schemas default parameter
DEFAULT_EVENT_SOURCE = "INSERT-YOUR-EVENT-SOURCE"
DEFAULT_EVENT_DETAIL_TYPE = "INSERT-YOUR-DETAIL-TYPE"

# page CLI constant
PAGE_LIMIT = 10
